unit modbus;

{$IFDEF FPC}
 {$mode objfpc}
 {$smartlink on}
{$ENDIF}
{$H+}

{ $DEFINE MODBUS_DEBUG}

interface

uses
  Classes, SysUtils, Sockets;

type

  TModbusFunction = (
    mbfReadCoilStatus=$01,
    mbfReadInputStatus=$02,
    mbfReadHoldingRegister=$03,
    mbfReadInputRegisters=$04
  );

  { TModbusClient }

  TModbusClient = class
    private
      Fsocket: TSocket;
      Fhost: string;
      FtransactionID: Word;
      function GetConnected: boolean;
      function GetNextTransactionID: Word;
      procedure Send(data: Pointer; len: integer);
      function Recv(data: Pointer; len: integer): boolean;
      function SendRecv(FunctionCode: TModbusFunction; RegisterNo,RegisterCount: Word; out Values: array of Word): boolean;
    public
      constructor Create;
      destructor Destroy; override;
      procedure Connect;
      procedure Disconnect;
      function ReadInputRegister(const RegisterNo: Word; out Value: Word): Boolean;
      function ReadInputRegisters(const RegisterNo,RegisterCount: Word; out Values: array of Word): Boolean;
      property Connected: boolean read GetConnected;
      property Host: string read Fhost write Fhost;
  end;

  { EModbusError }

  EModbusError = class(Exception)
    public
      ErrorCode: integer;
      constructor CreateFmt(const Error: integer; const Msg: string; const Args: array of const); overload;
  end;

implementation

const
  MODBUS_PORT = 502;
  MODBUS_PROTOCOL_ID = 0;
  MODBUS_ERRORS: array[1..6] of string = (
    'The function code received in the query is not an allowable action for the slave',   // ILLEGAL FUNCTION
    'The data address received in the query is not an allowable address for the slave',   // ILLEGAL DATA ADDRESS
    'A value contained in the query data fields is not an allowable value for the slave', // ILLEGAL DATA VALUE
    'The server failed during the execution',                                             // Server Failure
    'The server accepted the service invocation but the service requires a relatively long time to execute', // The  server  therefore  returns  only  an  acknowledgement  of  the  service invocation receipt.
    'The  server was unable to accept the MB Request PDU'                                 // Server busy
  );

type
  TModbusHeader = packed record
    TransactionID: word; // 1-2: 2 byte: Transaction identifier
    ProtocolID: word;    // 3-4: 2 byte: Protocol identifier (0 for Modbus/TCP)
    Length: word;        // 5-6: 2 byte: Length field (Number of remaining bytes in this frame)
    UnitID: byte;        // 7: 1 byte: Unit identifier (Slave address (255 if not used))
    FunctionCode: byte;  // 8: 1 byte: Function code
                         // (LRC or CRC-16 error check field is not required for Modbus TCP/IP)
  end;

  TModbusBuffer = packed record
    Header: TModbusHeader;
    Bytes: array[0..255] of Byte;
  end;

{$IFDEF MODBUS_DEBUG}
procedure DebugLn(msg: string; data: Pointer; len: integer);
const DEBUG_FILE='modbus.log';
var LogFile: TextFile; s: string;
begin
  {$I-}
  system.assign(LogFile, DEBUG_FILE);
  if FileExists(DEBUG_FILE) then
     system.append(LogFile)
  else
     system.rewrite(LogFile);
  SetString(s, data, len);
  writeln(LogFile, Format('%s: (%d): %s', [msg, len, s]));
  system.close(LogFile);
  {$I+}
end;
{$ENDIF}

{ EModbusError }

constructor EModbusError.CreateFmt(const Error: integer; const Msg: string; const Args: array of const);
begin
  ErrorCode := Error;
  CreateFmt(Msg, Args);
end;

{ TModbusClient }

constructor TModbusClient.Create;
begin
  inherited;
  FSocket := TSocket(INVALID_SOCKET);
end;

destructor TModbusClient.Destroy;
begin
  Disconnect;
  inherited;
end;

procedure TModbusClient.Connect;
var
  addr: TSockAddr;
  opt: LongWord;
begin
   FSocket := fpsocket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
   if FSocket = TSocket(INVALID_SOCKET) then
      raise EModbusError.CreateFmt(SocketError, 'Chyba %d pri vytvarani socketu!', [SocketError]);

   addr.sin_family := AF_INET;
   addr.sin_addr := StrToNetAddr(FHost);
   addr.sin_port := htons(MODBUS_PORT);
   if fpconnect(FSocket, @addr, sizeof(addr)) = SOCKET_ERROR then
      raise EModbusError.CreateFmt(SocketError, 'Chyba %d pri spojeni socketu!'#13#10'%s', [SocketError, SysErrorMessage(SocketError)]);
   opt := 3000; // timeout in milliseconds (ak vyprsi timeout: chyba 10060 - Connection timed out)
   fpsetsockopt(FSocket, SOL_SOCKET, SO_RCVTIMEO, @opt, SizeOf(opt));
   {$IFDEF MODBUS_DEBUG}DebugLn('Connect',nil,0);{$ENDIF}
end;

procedure TModbusClient.Disconnect;
begin
   fpshutdown(FSocket, SHUT_RDWR);
   closesocket(FSocket);
   FSocket := TSocket(INVALID_SOCKET);
   {$IFDEF MODBUS_DEBUG}DebugLn('Disconnect',nil,0);{$ENDIF}
end;

procedure TModbusClient.Send(data: Pointer; len: integer);
begin
  {$IFDEF MODBUS_DEBUG}DebugLn('Send',data,len);{$ENDIF}
  if fpsend(FSocket, data, len, 0) = SOCKET_ERROR then
     raise EModbusError.CreateFmt(SocketError, 'Chyba %d pri odosielani dat!'#13#10'%s', [SocketError, SysErrorMessage(SocketError)]);
end;

function TModbusClient.GetConnected: boolean;
begin
  Result := FSocket <> TSocket(INVALID_SOCKET);
end;

function TModbusClient.GetNextTransactionID: Word;
begin
  if FtransactionID=$FFFF then
     FtransactionID:=0
  else
    Inc(FtransactionID);
  Result := FtransactionID;
end;

function TModbusClient.Recv(data: Pointer; len: integer): boolean;
begin
  Result := len = fprecv(FSocket, data, len, 0);
  if not Result then
     // 10060 - Connection timed out - connected party did not properly respond after a period of time
     raise EModbusError.CreateFmt(SocketError, 'Chyba %d pri prijme dat!'#13#10'%s', [SocketError, SysErrorMessage(SocketError)]);
  {$IFDEF MODBUS_DEBUG}DebugLn('Recv',data,len);{$ENDIF}
end;

function TModbusClient.SendRecv(FunctionCode: TModbusFunction; RegisterNo, RegisterCount: Word; out Values: array of Word): boolean;
var
  Buffer: TModbusBuffer;
  Count,i: Byte;
begin
  // It is recommended to keep the TCP connection opened with a remote device and not to open and close it for each MODBUS/TCP transaction,
  // Several  MODBUS  transactions  can  be  activated  simultaneously  on  the  same  TCP  Connection.
  //  Remark:  If  this  is  done  then  the  MODBUS  transaction  identifier  must  be  used  to  uniquely identify the matching requests and responses.
  if not Connected then Connect;
  Buffer.Header.TransactionID := NtoBE(GetNextTransactionID); // It is used for transaction pairing, the MODBUS server copies in the response the transaction identifier of the request.
  Buffer.Header.ProtocolID := MODBUS_PROTOCOL_ID;
  Buffer.Header.UnitID := 0;
  Buffer.Header.FunctionCode := ord(FunctionCode);
  case FunctionCode of
    mbfReadInputRegisters:
      begin
        // Read the binary contents of input registers in the slave.
        //  Function Code 4 (read Analog Input Registers)
        //  2 bytes: The Data Address of the first register requested
        //  2 bytes: The total number of registers requested
        Buffer.Bytes[0] := Hi(RegisterNo);
        Buffer.Bytes[1] := Lo(RegisterNo);
        Buffer.Bytes[2] := Hi(RegisterCount);
        Buffer.Bytes[3] := Lo(RegisterCount);
        Buffer.Header.Length := NtoBE(Word(6)); // first byte (MSB)=0, second byte (LSB)=Length
      end;
  end;
  Send(@Buffer, 6+BEtoN(Buffer.Header.Length));
  Recv(@Buffer, 6);
  Recv(@Buffer.Header.UnitID, BEtoN(Buffer.Header.Length));
  if Buffer.Header.FunctionCode = ord(FunctionCode) then begin
    case FunctionCode of
      mbfReadInputRegisters:
        begin
          Count := Buffer.Bytes[0] div 2;
          i := 0;
          while (Count>0) and (i<Length(Values)) do begin
            Values[i] := BEtoN(PWord(@Buffer.Bytes[1+2*i])^);
            Inc(i);
            Dec(Count);
          end;
          Result := Length(Values)>0;
        end
      else
        Result := False;
    end;
  end
  else if Buffer.Header.FunctionCode = $80 or ord(FunctionCode) then begin
    // If the slave receives the query without a communication error, but cannot handle it (for example, if the request is to read anon-existent coil or register),
    //  the slave will return an exception response informing the master of the nature of the error
    // In an exception response, the slave echoes the slave address in the query,
    //  and sets the most-significant bit of the function code to 1.
    i := Buffer.Bytes[0];
    if (i>=Low(MODBUS_ERRORS)) and (i<=High(MODBUS_ERRORS)) then
       raise EModbusError.CreateFmt(i, MODBUS_ERRORS[i], [])
    else
       raise EModbusError.CreateFmt(0, 'Modbus protocol error No.%d!',[i]);
    Result := False;
  end
  else begin
    // If the slave does not receive the query due to a communications error, no response is returned.
    //  The master program will eventually process a timeout condition for the query.
    Disconnect;
    Result := False;
  end;
end;

function TModbusClient.ReadInputRegister(const RegisterNo: Word; out Value: Word): Boolean;
var Values: array[0..0] of Word;
begin
  Result := ReadInputRegisters(RegisterNo, 1, Values);
  if Result then Value := Values[0];
end;

function TModbusClient.ReadInputRegisters(const RegisterNo,RegisterCount: Word; out Values: array of Word): Boolean;
begin
  Result := SendRecv(mbfReadInputRegisters, RegisterNo, RegisterCount, Values);
end;

end.

